#include <stdio.h>

#include <sys/time.h>

#include "tsp.h"

#include "dsm.h"

#define DSM_SEG_ID    14
#define MAX_MACHINES  20

static int     numprocs;                /* total number of processors */
int            nodeid;                  /* id of this node            */
static int     master=0;                /* master or slave?           */

static char    pagemanager_host[40];
static char    barriermanager_host[40];
static char    lockmanager_host[40];

/* -- Used by the master program only */
static char  * machines[MAX_MACHINES];  /* names of participating machines */
static int     n_machines;              /* number of such machines         */
static struct timeval	start, end;     /* measure the time elapsed        */


static void read_machines() {

   /* Reads the machine names specified in file 'dsmtesthosts'.
   Updates variables 'machines' and 'n_machines'. */

    FILE * f;
    int    i;
    int    quit;
    char * p;
    int    l;

    if ((f = fopen("dsmtesthosts", "r")) < 0)
        perror("read_machines: fopen");

    for(i=0; i<MAX_MACHINES; i++) {
           machines[i] = (char*)malloc(40);
    }

    n_machines = 0;
    quit       = 0;

    printf("Going to read machines\n");

    while ((!quit) && (n_machines < MAX_MACHINES)) {
           p = fgets(machines[n_machines], 40, f);
        if (p) {
               l = strlen(machines[n_machines]);
            machines[n_machines][l-1] = 0;
            printf("read machine : %s\n",machines[n_machines]);
            if (l > 1)
                n_machines++;
        }
        else
            quit = 1;
    }

    printf("read machines\n");

    fclose(f);
}


#define SILENT

/* BEGIN SHARED DATA */
GlobalMemory	*global = NULL;
/* END SHARED DATA */


int 		StartNode, TspSize, NodesFromEnd;
int 		dump, debug, debugPrioQ;
int		performance;
extern int	visitNodes;

/* to print result file */
char 	        fname[50];
char            resfile[50]="result0";
FILE            *resfp;

usage()
{
   fprintf(stderr, "tsp:\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\n",
		"-M: indicates master or not  $ optional $",
		"-n: number of processes",
		"-i: node id $ optional for master - default = 1 $",
		"-P: page manager host",
		"-B: barrier manager host",
		"-L: lock manager host",
		"-f: size (filename)  $ optional $");

   exit(0);
}

void initialize()
{
   char the_command[400];
   int i;

   nodeid = 1;
   memset(global, 0, sizeof(*global));
   global->TourStackTop = -1;
   global->MinTourLen = BIGINT;
   global->TourLock = 1;
   global->MinLock = 2;
   global->barrier = 0;

   /* Read in the data file with graph description. */
   TspSize = read_tsp(fname);

   /* Initialize the first tour. */
   global->Tours[0].prefix[0] = StartNode;
   global->Tours[0].conn = 1;
   global->Tours[0].last = 0;
   global->Tours[0].prefix_weight = 0;
   calc_bound(0);	/* Sets lower_bound. */

   /* Initialize the priority queue structures. */
   global->PrioQ[1].index = 0;	/* The first PrioQ entry is Tour 0. */
   global->PrioQ[1].priority = global->Tours[0].lower_bound;
   global->PrioQLast = 1;

   /* Put all the unused tours in the free tour stack. */
   for (i = MAX_NUM_TOURS - 1; i > 0; i--)
      global->TourStack[++global->TourStackTop] = i;


   initBarrier(BARRIER_ID,numprocs);

   initLock(TOURLOCK);
   initLock(MINLOCK);

   read_machines();

   printf("start other processes...");
   printf("numprocs=%d...\n",numprocs);

   for (i=0; i<(numprocs-1); i++) 
   {
      sprintf(the_command, "rsh -n %s tsp -n %d -i %d -P %s -B %s -L %s&",
      machines[i], numprocs, i+2, pagemanager_host, barriermanager_host,machines[i],lockmanager_host);

      printf("execute command   %s\n", the_command);
      system(the_command);
    }

    printf("done\n");
}

extern char    *optarg;

main(argc,argv)
unsigned argc;
char **argv;
{
    int 		c, silent;

    NodesFromEnd = 12;
    dump = 0;

#ifdef SILENT
    silent = 1;
#endif /*SILENT*/

    /* solves the 8-node problem by default */
    strcpy(fname,"8");
    TspSize = 8;

    while ((c = getopt(argc, argv, "n:i:P:B:L:f:M")) != -1)
       switch (c) {
          case 'P' : 
		     sprintf(pagemanager_host, "%s", optarg);
                     break;
          case 'B' : 
		     sprintf(barriermanager_host, "%s", optarg);
                     break;
          case 'n' : 
		     numprocs = atoi(optarg);
                     break;
          case 'i' : 
		     nodeid = atoi(optarg);
		     sprintf(resfile,"result%d",nodeid);
                     break;
          case 'L' : 
		     sprintf(lockmanager_host, "%s", optarg);
                     break;
	  case 'f':
	 	     strcpy(fname, optarg);
		     TspSize = atoi(fname);
		     break;
          case 'M' : 
		     master = 1;
                     break;
	  case '?':
	  case 'h':
	  	     usage();
        }

    printf("numprocs = %d\n", numprocs);

    initDSM(/* you will have to put some arguments here */);

    global = (GlobalMemory *)openDSMSegment(DSM_SEG_ID,
                                        sizeof(GlobalMemory), master);

    if(master)
       initialize();

    resfp = fopen(resfile,"w");

    gettimeofday(&start, NULL);

    Worker();

    terminateDSM();

    fclose(resfp);

}


void printresultfile()
{
    int i;
    fprintf(resfp, "########################################\n\n");
    fprintf(resfp, "\t[%d: MINIMUM TOUR LENGTH: %d]\n\n",nodeid,global->MinTourLen);
    fprintf(resfp, "%d: MINIMUM TOUR:\n",nodeid);
	for (i = 0; i < TspSize; i++) {
	    if (i % 10) fprintf(resfp, " - ");
	    else fprintf(resfp, "\n\t");
	    fprintf(resfp, "%2d", (int) global->MinTour[i]);
	}
    fprintf(resfp, " - %2d\n\n",StartNode);
    fprintf(resfp, "########################################\n\n");
}


Worker()
{
    int curr = -1;

    int ind=0;

    int i;

    waitAtBarrier(BARRIER_ID);   /* Wait for signal to start. */

    for (;;) {
	/* Continuously get a tour and split. */
	curr = get_tour(curr);
	if (curr < 0) {
	    break;
	}
	recursive_solve(curr);
    }

    waitAtBarrier(BARRIER_ID);	 /* Signal completion. */

    printresultfile();

    if(master)
    {

    gettimeofday(&end, NULL);

    fprintf(stderr, "########################################\n\n");

    fprintf(stderr, "Elapsed time: %.2f seconds\n",
	    (((end.tv_sec * 1000000.0) + end.tv_usec) -
	     ((start.tv_sec * 1000000.0) + start.tv_usec)) / 1000000.0);

    fprintf(stderr, "\n-----------------\n\n");

    fprintf(stderr, "\t[MINIMUM TOUR LENGTH: %d]\n\n", global->MinTourLen);
    fprintf(stderr, "MINIMUM TOUR:\n");
	for (i = 0; i < TspSize; i++) {
	    if (i % 10) fprintf(stderr, " - ");
	    else fprintf(stderr, "\n\t");
	    fprintf(stderr, "%2d", (int) global->MinTour[i]);
	}
    fprintf(stderr, " - %2d\n\n", StartNode);

    fprintf(stderr, "########################################\n\n");

    }

    waitAtBarrier(BARRIER_ID);   /* finished Worker */
}
